# import logging
#
# # Dictionary with the error types
# error_message_dict = {}
# from sys import exit
#
#
# class LimeLogger:
#
#     def __init__(self):
#
#         self.messages = []
#
#     def note(self, message):
#
#         self.messages.append(message)
#
#     def inform(self, message):
#
#         print(message)
#         self.messages.append(message)
#
#     def message(self, message):
#
#         print(message)
#         self.messages.append(message)
#
#         return
#
#     def warning(self, message):
#
#         print(message)
#         self.messages.append(message)
#
#         return
#
#     def bug(self, message):
#
#         print(message)
#         self.messages.append(message)
